/*
 *	SCCS: @(#)scen1.c	1.7 (98/03/05)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)scen1.c	1.7 (98/03/05) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)scen1.c	1.7 98/03/05 TETware release 3.2
NAME:		scen1.c
PRODUCT:	TETware
AUTHOR:		Andrew Dingwall, UniSoft Ltd.
DATE CREATED:	August 1996

DESCRIPTION:
	scenario parser stage 1 - tokenise the scenario file

MODIFICATIONS:
	Andrew Dingwall, UniSoft Ltd., March 1997
	allow leading white space on lines read from a scenario include file

	Andrew Dingwall, UniSoft Ltd., March 1998
	Added support for number ranges in directive arguments.
	Increased the number of arguments that a directive may take.
	Permit a number range to appear as an argument to a remote or
	distributed directive, so as to avoid the need to handle long
	scenario lines and large numbers of arguments when many system
	IDs are specified.

************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include "dtmac.h"
#include "error.h"
#include "ltoa.h"
#include "scentab.h"
#include "dirtab.h"
#include "dtetlib.h"
#include "tcc.h"


#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;	/* file name for error reporting */
#endif

/* some limits which apply to directives */
/* max number of semicolon-delimited fields in group execution */
#define MAXFLDS	(LBUFLEN / 8)
#if MAXFLDS < 10
#define MAXFLDS	10
#endif
/* max number of comma-delimited args in a single directive */
#define MAXARGS	(LBUFLEN / 4)
#if MAXARGS < 10
#define MAXARGS	10
#endif

/* is this an attached test list or file name? */
#define isattached(s)	(*(s) && !isspace(*(s)))

/* static function declarations */
static int find1scen PROTOLIST((FILE *, char *, int *));
static void includefile PROTOLIST((char *, char *, int));
static int isnumrange PROTOLIST((char *, int *, int *));
static int iszorpnum PROTOLIST((char *));
static int proc1dgrp PROTOLIST((char *, char *, char *, int));
static void proc1scelem PROTOLIST((char *, int, int, int, char *));
static int proc1scen PROTOLIST((FILE *, char *, int *));
static int proc1scline PROTOLIST((char *, char *, int));


/*
**	proc1scfile() - tokenise a scenario input file
**
**	return 0 if successful or -1 on error
*/

int proc1scfile(fp, fname)
FILE *fp;
char *fname;
{
	int rc1, rc2;
	int lcount = 0;

	TRACE2(tet_Tscen, 1, "proc1scfile(): tokenising scenario file %s",
		fname);

	/* tokenise each scenario in turn */
	for (;;) {
		if ((rc1 = find1scen(fp, fname, &lcount)) > 0) {
			while ((rc2 = proc1scen(fp, fname, &lcount)) != 0)
				if (rc2 < 0)
					return(rc2);
		}
		else
			return(rc1);
	}
}

/*
**	find1scen() - search for the start of a scenario in the input file
**
**	return	 1 if found
**		 0 on EOF
**		-1 if processing of this file must be abandoned
*/

static int find1scen(fp, fname, lcp)
FILE *fp;
char *fname;
int *lcp;
{
	char *line;
	int skipstart = 0;
	static char fmt[] = "found a scenario after skipping %d line%s";
	char msg[sizeof fmt + LNUMSZ];

	TRACE1(tet_Tscen, 4, "find1scen(): find start of scenario");

	/* find the start of the next scenario -
		skip lines until it is found -
		when the start is found, push the line back so that
		the scenario processor can find it */
	for (;;) {
		if ((line = getline(fp, fname, lcp)) == (char *) 0)
			return(ferror(fp) ? -1 : 0);
		if (!isspace(*line)) {
			/* start of new scenario */
			if (skipstart) {
				(void) sprintf(msg, fmt, *lcp - skipstart,
					*lcp - skipstart == 1 ? "" : "s");
				scenermsg(msg, (char *) 0, *lcp, fname);
			}
			ungetline(fp, line, *lcp);
			return(1);
		}
		/* here if we are skipping lines -
			emit a warning first time through */
		if (!skipstart) {
			skipstart = *lcp;
			scenerror("line(s) outside of a scenario ignored",
				(char *) 0, skipstart, fname);
		}
	}
}

/*
**	proc1scen() - tokenise a scenario once the input stream is
**		correctly positioned
**
**	return	 1 if successful and the input stream is still
**			correctly positioned
**		 0 if successful and the input stream is not
**			correctly positioned (in practice this means EOF)
**		-1 if processing of this file must be abandoned
*/

static int proc1scen(fp, fname, lcp)
FILE *fp;
char *fname;
int *lcp;
{
	char *line, *next;
	register char *p;
	register struct scentab *ep;

	/* read the scenario name - starts in column 1 */
	line = getline(fp, fname, lcp);
	ASSERT(line);
	ASSERT(!isspace(*line));

	/* see if there are scenario elements after the name */
	for (p = line; *p; p++)
		if (isspace(*p)) {
			*p++ = '\0';
			break;
		}
	next = p;

	TRACE3(tet_Tscen, 4, "proc1scen(): new scenario = <%s>, rest = <%s>",
		line, next);

	/* store the scenario header token */
	ep = scalloc();
	ep->sc_type = SC_SCENARIO;
	ep->sc_scenario = rstrstore(line);
	ep->sc_lineno = *lcp;
	ep->sc_fname = fname;
	scpush(ep, &sclist);

	/* process any elements on the same line as the header */
	if (proc1scline(next, fname, *lcp) < 0)
		return(-1);

	/* process the rest of the current scenario */
	while ((line = getline(fp, fname, lcp)) != (char *) 0) {
		if (!isspace(*line)) {
			/* a new scenario */
			ungetline(fp, line, *lcp);
			return(1);
		}
		/* process a line in the current scenario */
		if (proc1scline(line, fname, *lcp) < 0)
			return(-1);
	}

	return(ferror(fp) ? -1 : 0);
}

/*
**	proc1cmdline() - tokenise a scenario line specified with
**		the -l command-line option
**
**	return 0 if successful or -1 if processing must be abandoned
*/

int proc1cmdline(line)
char *line;
{
	static char cmdline[] = "<command-line>";
	static int lineno;
	struct scentab *ep;

	/* start a new (default) scenario first time through */
	if (!lineno++) {
		ep = scalloc();
		ep->sc_type = SC_SCENARIO;
		ep->sc_scenario = "all";
		ep->sc_lineno = lineno;
		ep->sc_fname = cmdline;
		scpush(ep, &sclist);
	}

	/* process the line and return */
	return(proc1scline(line, cmdline, lineno));
}

/*
**	proc1scline() - tokenise a line in the current scenario
**
**	return 0 if successful or -1 if processing of this file must be
**		abandoned
*/

static int proc1scline(line, fname, lineno)
char *line, *fname;
int lineno;
{
	register char *p;
	int type, flags;
	char *next;
	char delim;
	static char msg[2];

	/* skip leading spaces */
	while (*line && isspace(*line))
		line++;

	TRACE2(tet_Tscen, 4, "proc1scline(): line = <%s>", line);

	/* determine the element type and how it is delimited */
	flags = 0;
	switch (*line) {
	case '\0':	/* an empty line */
		return(0);
	case '"':	/* a scenario information line */
		type = SC_SCENINFO;
		delim = *line++;
		break;
	case '^':	/* a referenced scenario name */
		type = SC_SCEN_NAME;
		delim = ' ';
		line++;
		break;
	case ':':	/* a scenario directive */
		type = SC_DIRECTIVE;
		delim = *line++;
		break;
	default:	/* treat everything else as a testcase for now -
				we'll check for valid testcase names later */
		type = SC_TESTCASE;
		delim = ' ';
		break;
	}

	/* search for the required delimiter */
	for (p = line; *p; p++)
		if (*p == delim || (delim == ' ' && isspace(*p)))
			break;

	/* see if it was found */
	if (!*p && !isspace(delim)) {
		msg[0] = delim;
		scenerror("unmatched delimiter", msg, lineno, fname);
		return(0);
	}

	/* terminate this element, remember start of next element */
	if (*p)
		*p++ = '\0';
	next = p;

	TRACE3(tet_Tscen, 4, "proc1scline(): element = <%s>, rest = <%s>",
		line, next);

	/* process this element */
	switch (type) {
	case SC_TESTCASE:
	case SC_SCEN_NAME:
		if (!*(line + 1)) {
			scenerror("empty test case or scenario name",
				line, lineno, fname);
			break;
		}
		/* else fall through */
	case SC_SCENINFO:
		proc1scelem(line, type, flags, lineno, fname);
		break;
	case SC_DIRECTIVE:
		return(proc1dgrp(line, next, fname, lineno));
	default:
		/* this "can't happen" .*/
		fatal(0, "unexpected type", prsctype(type));
		/* NOTREACHED */
	}

	/* process the rest of the current line */
	return(proc1scline(next, fname, lineno));
}

/*
**	proc1scelem() - process a single (non-directive) scenario element
*/

static void proc1scelem(element, type, flags, lineno, fname)
char *element, *fname;
int type, flags, lineno;
{
	register struct scentab *ep;
	register char *p;
	char *iclist, *tcname;
	int lsceninfo;

	TRACE2(tet_Tscen, 4, "proc1scelem(): element = <%s>", element);

	/* allocate a scenario element and fill it in */
	ep = scalloc();
	ep->sc_type = type;
	ep->sc_flags = flags;
	ep->sc_lineno = lineno;
	ep->sc_fname = fname;

	switch (type) {
	case SC_SCENINFO:
		lsceninfo = 0;
		RBUFCHK(&ep->sc_sceninfo, &lsceninfo,
			(int) strlen(element) + 3);
		(void) sprintf(ep->sc_sceninfo, "\"%s\"", element);
		break;
	case SC_TESTCASE:
		tcname = element;
		iclist = (char *) 0;
		for (p = element; *p; p++)
			if (*p == '{') {
				*p++ = '\0';
				for (iclist = p; *p; p++)
					if (*p == '}') {
						*p = '\0';
						break;
					}
				break;
			}
		ep->sc_tcname = rstrstore(tcname);
		if (iclist && *iclist)
			ep->sc_sciclist = rstrstore(iclist);
		ep->sc_exiclist = ep->sc_sciclist;
		break;
	case SC_SCEN_NAME:
		ep->sc_scen_name = rstrstore(element);
		break;
	default:
		/* this "can't happen" */
		fatal(0, "unexpected type", prsctype(type));
		/* NOTREACHED */
	}

	/* then add the element to the list */
	scpush(ep, &sclist);
}

/*
**	proc1dgrp() - process a group of scenario directives;
**		i.e., something between a pair of ':' delimiters
**
**	return 0 if successful or -1 if processing of this file must be
**		abandoned
*/

static int proc1dgrp(dgroup, next, fname, lineno)
char *dgroup, *next, *fname;
int lineno;
{
	char buf[LBUFLEN];
	char *dirs[MAXFLDS];
	int ndirs;
	char **dirp;
	char *args[MAXARGS];
	int nargs;
	register char **ap;
	register struct scentab *ep;
	register struct dirtab *dp;
	char **vp;
	char *line;
	int istart, iend, ok, rc;
	register char *p1, *p2;
	struct scentab *scstack = (struct scentab *) 0;
#ifndef TET_LITE	/* -START-LITE-CUT- */
	int *ip, nsys, syslen;
#endif /* !TET_LITE */	/* -END-LITE-CUT- */

	TRACE3(tet_Tscen, 4,
		"proc1dgrp(): directive group = <%s>, rest = <%s>",
		dgroup, next);

	/* now the colons have gone - split a group-execution directive
		into semicolon-delimited fields */
	ndirs = split(dgroup, dirs, MAXFLDS, ';');

	/* trim leading spaces from each directive */
	for (dirp = dirs; dirp < &dirs[ndirs]; dirp++)
		while (**dirp && isspace(**dirp))
			++*dirp;

#ifndef NOTRACE
	TRACE2(tet_Tscen, 4, "group contains %s directive(s)", tet_i2a(ndirs));
	if (tet_Tscen >= 5)
		for (dirp = dirs; dirp < &dirs[ndirs]; dirp++)
			TRACE3(tet_Tscen, 5, "\tdirective %s = <%s>",
				tet_i2a((dirp - dirs) + 1), *dirp);
#endif

	/*
	**	check the syntax of the directive group
	*/

#ifndef NOTRACE
	if (isattached(next))
		TRACE2(tet_Tscen, 4,
			"directive group has attached items: <%s>",
			next);
	else
		TRACE1(tet_Tscen, 4, "directive group has no attached items");
#endif

	/* check the syntax of each directive in the group
		using the rules in the directives table */
	TRACE1(tet_Tscen, 4, "starting directive syntax check");
	ok = 1;
	for (dirp = dirs; dirp < &dirs[ndirs]; dirp++) {
		/* count the number of comma-separated arguments
			associated with this directive */
		(void) strcpy(buf, *dirp);
		args[0] = "";
		nargs = split(buf, args, MAXARGS, ',') - 1;

		/* here, args[0] is the directive,
			nargs is the number of arguments
			args[1] -> args[nargs] are the arguments */

		/* trim leading spaces from each argument */
		for (ap = &args[1]; ap <= &args[nargs]; ap++)
			while (**ap && isspace(**ap))
				++*ap;

#ifndef NOTRACE
		TRACE3(tet_Tscen, 6,
			"syntax checking <%s> with %s argument(s)",
			args[0], tet_i2a(nargs));
		if (tet_Tscen >= 7)
			for (ap = &args[1]; ap <= &args[nargs]; ap++)
				TRACE3(tet_Tscen, 7, "\targ %s = <%s>",
					tet_i2a(ap - args), *ap);
#endif

		/* look up the directive in the table */
		if ((dp = getdirbyname(args[0])) == (struct dirtab *) 0) {
			if (!*dirp)
				*dirp = "<empty>";
			scenerror("unknown/unsupported directive",
				*dirp, lineno, fname);
			ok = 0;
			continue;
		}

		/* check for optional and mandatory arguments */
		if (
			nargs == 0 &&
			(dp->dt_flags & (SDF_NEED_ARG | SDF_NEED_MARG))
		) {
			scenerror(args[0], "needs an argument",
				lineno, fname);
			ok = 0;
		}
		else if (
			nargs > 1 &&
			(dp->dt_flags & (SDF_NEED_ARG | SDF_OPT_ARG))
		) {
			scenerror(args[0], "has too many arguments",
				lineno, fname);
			ok = 0;
		} else if (
			nargs > 0 &&
			!(dp->dt_flags & (SDF_OPT_ARG | SDF_OPT_MARG |
				SDF_NEED_ARG | SDF_NEED_MARG))
		) {
			scenerror(args[0], "should not have arguments",
				lineno, fname);
			ok = 0;
		}

		/* check for non-numeric arguments */
		if (dp->dt_flags & SDF_NUMERIC_ARGS)
			for (ap = &args[1]; ap <= &args[nargs]; ap++)
				if (!iszorpnum(*ap)) {
					scenerror(args[0],
						"argument is not a +ve number",
						lineno, fname);
					ok = 0;
					break;
				}

		/* check for correctly formatted number range arguments */
		if (dp->dt_flags & SDF_NRANGE_ARGS)
			for (ap = &args[1]; ap <= &args[nargs]; ap++) {
				istart = iend = 0;
				if (
					!iszorpnum(*ap) &&
					!isnumrange(*ap, &istart, &iend)
				) {
					scenerror(args[0],
			"argument is not a +ve number or number range",
						lineno, fname);
					ok = 0;
					break;
				}
				else if (istart > iend) {
					scenerror(*ap,
			": start is greater than end in number range",
						lineno, fname);
					ok = 0;
				}
			}

		/* check for "variable" format arguments */
		if (dp->dt_flags & SDF_VARFMT_ARGS)
			for (ap = &args[1]; ap <= &args[nargs]; ap++) {
				p1 = tet_remvar(*ap, -1);
				p2 = tet_equindex(*ap);
				if (!p1 || !p2) {
					scenerror(*ap,
			"is not a properly formatted variable assignment",
						lineno, fname);
					ok = 0;
					break;
				}
				else if (!strncmp(p1, "TET_", 4)) {
					*p2 = '\0';
					scenerror(p1,
			"value may not be assigned in a scenario file",
						lineno, fname);
					ok = 0;
					break;
				}
			}

		/* check for optional and mandatory attached elements */
		if (isattached(next)) {
			if ((dp->dt_flags & (SDF_OPT_ATTACH | SDF_NEED_ATTACH)) == 0) {
				scenerror(args[0],
					"should not have an attached element",
					lineno, fname);
				ok = 0;
			}
		}
		else {
			if (dp->dt_flags & SDF_NEED_ATTACH) {
				scenerror(args[0],
					"needs an attached element",
					lineno, fname);
				ok = 0;
			}
		}
	}

	TRACE2(tet_Tscen, 6, "directive syntax check was %ssuccessful",
		ok ? "" : "un");

	if (!ok) {
		TRACE1(tet_Tscen, 4, "proc1dgrp(): return 0");
		return(0);
	}

	/* process each of the directives in turn */
	for (dirp = dirs; dirp < &dirs[ndirs]; dirp++) {

		TRACE2(tet_Tscen, 4, "process directive <%s>", *dirp);

		/* separate the directive from its arguments */
		nargs = split(*dirp, args, MAXARGS, ',') - 1;

		/* trim leading spaces from each argument */
		for (ap = &args[1]; ap <= &args[nargs]; ap++)
			while (**ap && isspace(**ap))
				++*ap;

		/* here, args[0] is the directive,
			nargs is the number of arguments
			args[1] -> args[nargs] are the arguments */

#ifndef NOTRACE
		TRACE3(tet_Tscen, 4, "directive = <%s> has %s argument(s)",
			args[0], tet_i2a(nargs));
		if (tet_Tscen >= 5)
			for (ap = &args[1]; ap <= &args[nargs]; ap++)
				TRACE3(tet_Tscen, 7, "\targ %s = <%s>",
					tet_i2a(ap - args), *ap);
#endif

		/* look the directive up in the table */
		dp = getdirbyname(args[0]);
		ASSERT(dp);

		/* an INCLUDE directive does not go in the list -
		**	in dtet mode it is just a tag on which
		**	to hang a filename;
		**	in etet mode it is a no-op
		*/
		if (dp->dt_directive == SD_INCLUDE)
			continue;

		/* allocate a scenario element and fill it in */
		ep = scalloc();
		ep->sc_type = SC_DIRECTIVE;
		ep->sc_directive = dp->dt_directive;
		ap = &args[1];
		switch (ep->sc_directive) {
		case SD_PARALLEL:
		case SD_REPEAT:
			ASSERT(nargs == 0 || nargs == 1);
			if (nargs == 0)
				ep->sc_count = 1;
			else
				ep->sc_count = atoi(*ap);
			break;
		case SD_TIMED_LOOP:
			ASSERT(nargs == 1);
			ep->sc_seconds = atol(*ap);
			break;
		case SD_VARIABLE:
			ASSERT(nargs >= 0);
			if ((ep->sc_vars = (char **) calloc((unsigned) MAX(nargs, 1), sizeof *ep->sc_vars)) == (char **) 0)
				fatal(errno, "can't allocate memory for variable list", (char *) 0);
			TRACE2(tet_Tbuf, 6, "allocate variable list = %s",
				tet_i2x(ep->sc_vars));
			ep->sc_nsys = nargs;
			for (vp = ep->sc_vars; nargs > 0; nargs--)
				*vp++ = rstrstore(*ap++);
			errno = 0;
			break;
#ifndef TET_LITE	/* -START-LITE-CUT- */
		case SD_REMOTE:
		case SD_DISTRIBUTED:
			ASSERT(nargs > 0);
			ep->sc_nsys = 0;
			syslen = 0;
			for (; nargs > 0; nargs--, ap++) {
				nsys = 1;
				if (isnumrange(*ap, &istart, &iend))
					nsys += iend - istart;
				else
					istart = atoi(*ap);
				RBUFCHK((char **) &ep->sc_sys, &syslen, (int) (sizeof *ep->sc_sys * (ep->sc_nsys + nsys)));
				ip = ep->sc_sys + ep->sc_nsys;
				ep->sc_nsys += nsys;
				while (--nsys >= 0)
					*ip++ = istart++;
			}
			break;
#endif /* !TET_LITE */	/* -END-LITE-CUT- */
		case SD_RANDOM:
		case SD_END_PARALLEL:
		case SD_END_REPEAT:
		case SD_END_RANDOM:
		case SD_END_TIMED_LOOP:
		case SD_END_VARIABLE:
#ifndef TET_LITE	/* -START-LITE-CUT- */
		case SD_END_REMOTE:
		case SD_END_DISTRIBUTED:
#endif /* !TET_LITE */	/* -END-LITE-CUT- */
			ASSERT(nargs == 0);
			break;
		default:
			/* this "can't happen" */
			fatal(0, "unexpected directive",
				prscdir(dp->dt_directive));
			/* NOTREACHED */
		}
		ep->sc_fname = fname;
		ep->sc_lineno = lineno;

		/* store the directive */
		scpush(ep, &sclist);

		/*
		** if there is an attached element, we will need a matching
		** end* directive once the attached element is processed
		*/
		if (isattached(next) && (dp->dt_flags & SDF_ENDDIR)) {
			dp = getdirbyvalue(dp->dt_match);
			ASSERT(dp);
			ep = scalloc();
			ep->sc_type = SC_DIRECTIVE;
			ep->sc_directive = dp->dt_directive;
			ep->sc_flags = SCF_IMPLIED;
			ep->sc_fname = fname;
			ep->sc_lineno = lineno;
			scpush(ep, &scstack);
		}
	}

	/*
	** here we must decide what to do with the rest of the line -
	**
	** if the thing after the directive(s) is a space, there are no
	** implied end* directives and the rest of the line can be
	** processed separately
	*/

	if (!isattached(next)) {
		ASSERT(scstack == (struct scentab *) 0);
		TRACE1(tet_Tscen, 4,
			"no attached items - process rest of line");
		rc = proc1scline(next, fname, lineno);
		TRACE2(tet_Tscen, 4, "proc1dgrp(): RETURN %s", tet_i2a(rc));
		return(rc);
	}

	/*
	** but if the thing after the directive(s) is attached
	** we must determine the nature of the attached element
	** which could be one of the follwing:
	**
	**	/include-file-name
	**	@/test-case-name
	**	^scenario-name
	*/

	/* separate the attached element from the rest of the line */
	line = next;
	for (p1 = line; *p1; p1++)
		if (isspace(*p1)) {
			if (tet_compat == COMPAT_DTET)
				p2 = ":";
			else
				for (p2 = p1 + 1; *p2; p2++)
					if (!isspace(*p2))
						break;
			if (*p2 == ':') {
				*p1++ = '\0';
				break;
			}
		}
	next = p1;

	/*
	** now, line points to the (null-terminated) attached element
	** and next points to the rest of the line
	*/
	switch (*line) {
	case '@':
		line++;
		/* fall through */
	case '^':
		if (proc1scline(line, fname, lineno) < 0) {
			while ((ep = scpop(&scstack)) != (struct scentab *) 0)
				scfree(ep);
			TRACE1(tet_Tscen, 4, "proc1dgrp(): return -1");
			return(-1);
		}
		break;
	case '/':
		includefile(line, fname, lineno);
		break;
	default:
		scenerror("badly formed include file name:",
			line, lineno, fname);
		break;
	}

	/* add the implied end* directives */
	while ((ep = scpop(&scstack)) != (struct scentab *) 0)
		scpush(ep, &sclist);

	/* finally, process the rest of the line and return */
	TRACE1(tet_Tscen, 4, "process rest of line after attached items");
	rc = proc1scline(next, fname, lineno);
	TRACE2(tet_Tscen, 4, "proc1dgrp(): return %s", tet_i2a(rc));
	return(rc);
}

/*
**	includefile() - interpolate the contents of an include file
*/

static void includefile(nextfile, currfile, currline)
char *nextfile, *currfile;
int currline;
{
	char buf[MAX(LBUFLEN, MAXPATH)];
	FILE *fp;
	char *fname;
	int lineno = 0;
	char *args[1];
	int nargs;
	register char *p, *bp;
	char delim;

	TRACE2(tet_Tscen, 4, "includefile(): file = <%s>", nextfile);

	/* strip off unwanted characters */
	for (p = nextfile; *p; p++)
		if (!isdirsep(*p))
			break;
	if (!*p) {
		scenerror("empty include file name:", nextfile,
			currline, currfile);
		return;
	}

	/*
	** generate the full path name of the file relative to the
	** test suite root directory
	*/
	fullpath(tet_tsroot, p, buf, sizeof buf, 0);
	fname = rstrstore(buf);

	/* open the file */
	if ((fp = fopen(fname, "r")) == (FILE *) 0) {
		error(errno, "can't open", fname);
		scenerrors++;
		free(fname);
		return;
	}

	/* read each line in turn */
	while (fgets(buf, sizeof buf, fp) != (char *) 0) {
		lineno++;

		/* strip out comments and the trailing newline */
		for (p = buf; *p; p++)
			if (*p == '\n' || *p == '#') {
				*p = '\0';
				break;
			}
		/* strip trailing spaces */
		while (--p >= buf)
			if (isspace(*p))
				*p = '\0';
			else
				break;

		/* strip leading spaces */
		for (bp = buf; *bp; bp++)
			if (!isspace(*bp))
				break;

		TRACE2(tet_Tscen, 10, "line after stripping spaces = <%s>", bp);

		/* ignore a blank line */
		if (!*bp)
			continue;

		/* check that the line contains a valid element -
			only file names and scenario information lines
			are valid here
		*/
		switch (*bp) {
		case '/':
			delim = ' ';
			break;
		case '"':
			delim = '"';
			break;
		default:
			nargs = split(bp, args, 1, ' ');
			ASSERT(nargs >= 1);
			scenerror(args[0], "is not valid in included file",
				lineno, fname);
			continue;
		}

		/* ensure that there is only one element on the line */
		for (p = bp + 1; *p; p++)
			if (*p == delim || (delim == ' ' && isspace(*p))) {
				if (delim == '"')
					p++;
				if (split(p, args, 1, ' ') > 0) {
					scenerror("too many elements",
						(char *) 0, lineno, fname);
				}
				*p = '\0';
				break;
			}

		/* send the element for processing */
		if (proc1scline(bp, fname, lineno) < 0)
			break;
	}

	/* finally, close the file and return (but don't free fname!) */
	(void) fclose(fp);
	TRACE1(tet_Tscen, 4, "includefile(): return");
}

/*
**	iszorpnum() - see if a string is a non-negative number
**
**	return 1 if it is or 0 if it isn't
*/

static int iszorpnum(s)
register char *s;
{
	if (!*s)
		return(0);
	else
		while (*s && isdigit(*s))
			s++;

	return(*s ? 0 : 1);
}

/*
**      isnumrange() - see if a string is of the form m-n where m and n
**		are non-negative numbers
**
**	return 1 if it is or 0 if it isn't
**
**	if successful the values of m and n are returned indirectly
**		through *startp and *endp
*/

static int isnumrange(s, startp, endp)
char *s;
int *startp, *endp;
{
	register char *p;
	register int rc;

	for (p = s; *p; p++)
		if (*p == '-') {
			*p = '\0';
			if ((rc = iszorpnum(s) & iszorpnum(p + 1)) != 0) {
				if (startp)
					*startp = atoi(s);
				if (endp)
					*endp = atoi(p + 1);
			}
			*p = '-';
			return(rc);
		}

	return(0);
}

